package de.unima.ki.anyburl;

import java.io.File;
import java.io.FileInputStream;
import java.io.IOException;
import java.io.InputStream;
import java.io.PrintWriter;
import java.util.ArrayList;
import java.util.LinkedList;
import java.util.List;
import java.util.Properties;

import de.unima.ki.anyburl.data.TripleSet;
import de.unima.ki.anyburl.io.IOHelper;
import de.unima.ki.anyburl.io.RuleReader;
import de.unima.ki.anyburl.structure.Rule;
import de.unima.ki.anyburl.structure.RuleFilter;

public class Apply {
		
	
	private static String CONFIG_FILE = "config-apply-fb15.properties";
	
	
	/**
	 * Filter the rule set prior to applying it. Removes redundant rules which do not have any impact (or no desired impact). 
	 */
	public static boolean FILTER = true;	

	
	/**
	 * This should never be set to true, if set to true it activates a kind of trick, which might have been used by others
	 * implicitly or explicitly to get better values. This setting has only been used to check how strong the impact of this trick is.
	 */
	public static boolean FB15_237_TRICK = false;	
	
	
	public static void main(String[] args) throws IOException {
		
		if (args.length == 1) {
			CONFIG_FILE = args[0];
			System.out.println("reading params from file " + CONFIG_FILE);
		}
		
		Rule.applicationMode();
		Properties prop = new Properties();
		InputStream input = null;
		try {
			input = new FileInputStream(CONFIG_FILE);
			prop.load(input);
			Settings.PREDICTION_TYPE = IOHelper.getProperty(prop, "PREDICTION_TYPE", Settings.PREDICTION_TYPE);
			if (Settings.PREDICTION_TYPE.equals("aRx")) {
				Settings.PATH_TRAINING = IOHelper.getProperty(prop, "PATH_TRAINING", Settings.PATH_TRAINING);
				Settings.PATH_TEST = IOHelper.getProperty(prop, "PATH_TEST", Settings.PATH_TEST);
				Settings.PATH_VALID = IOHelper.getProperty(prop, "PATH_VALID",Settings.PATH_VALID);
				Settings.PATH_OUTPUT = IOHelper.getProperty(prop, "PATH_OUTPUT", Settings.PATH_OUTPUT);
				Settings.PATH_EXPLANATION = IOHelper.getProperty(prop, "PATH_EXPLANATION", Settings.PATH_EXPLANATION);
				Settings.PATH_RULES = IOHelper.getProperty(prop, "PATH_RULES", Settings.PATH_RULES);
				Settings.TOP_K_OUTPUT = IOHelper.getProperty(prop, "TOP_K_OUTPUT", Settings.TOP_K_OUTPUT);
				Settings.UNSEEN_NEGATIVE_EXAMPLES = IOHelper.getProperty(prop, "UNSEEN_NEGATIVE_EXAMPLES", Settings.UNSEEN_NEGATIVE_EXAMPLES);
				Settings.THRESHOLD_CONFIDENCE = IOHelper.getProperty(prop, "THRESHOLD_APPLIED_CONFIDENCE", Settings.THRESHOLD_CONFIDENCE);
				Settings.DISCRIMINATION_BOUND = IOHelper.getProperty(prop, "DISCRIMINATION_BOUND",  Settings.DISCRIMINATION_BOUND);
				Settings.TRIAL_SIZE = IOHelper.getProperty(prop, "TRIAL_SIZE",  Settings.TRIAL_SIZE);
				Settings.WORKER_THREADS = IOHelper.getProperty(prop, "WORKER_THREADS",  Settings.WORKER_THREADS);
				FILTER = IOHelper.getProperty(prop, "FILTER", FILTER);
				Settings.AGGREGATION_TYPE = IOHelper.getProperty(prop, "AGGREGATION_TYPE", Settings.AGGREGATION_TYPE);
				FB15_237_TRICK = IOHelper.getProperty(prop, "FB15_237_TRICK", FB15_237_TRICK);
				
				if (Settings.AGGREGATION_TYPE.equals("maxplus")) Settings.AGGREGATION_ID = 1;
				if (Settings.AGGREGATION_TYPE.equals("noisyor")) Settings.AGGREGATION_ID = 2;
			}
			/*
			else if (PREDICTION_TYPE.equals("xRy")) {
				TRESHOLD_FUNCTIONAL = IOHelper.getProperty(prop, "TRESHOLD_FUNCTIONAL", TRESHOLD_FUNCTIONAL);
				System.out.println("TF: " + TRESHOLD_FUNCTIONAL);
				PATH_TRAINING = IOHelper.getProperty(prop, "PATH_TRAINING",PATH_TRAINING);
				PATH_TEST = IOHelper.getProperty(prop, "PATH_TEST", PATH_TEST);
				PATH_VALID = IOHelper.getProperty(prop, "PATH_VALID",PATH_VALID);
				PATH_OUTPUT = IOHelper.getProperty(prop, "PATH_OUTPUT", PATH_OUTPUT);
				PATH_RULES = IOHelper.getProperty(prop, "PATH_RULES", PATH_RULES);
				TOP_K_OUTPUT = IOHelper.getProperty(prop, "TOP_K_OUTPUT", TOP_K_OUTPUT);
				COMBINATION_RULE = IOHelper.getProperty(prop, "COMBINATION_RULE", COMBINATION_RULE);
				if (COMBINATION_RULE.equals("multiplication")) RuleEngine.COMBINATION_RULE_ID = 1;
				if (COMBINATION_RULE.equals("maxplus")) RuleEngine.COMBINATION_RULE_ID = 2;
				if (COMBINATION_RULE.equals("max")) RuleEngine.COMBINATION_RULE_ID = 3;
			}
			*/
			else {
				System.err.println("The prediction type " + Settings.PREDICTION_TYPE + " is not yet supported.");
				System.exit(1);
			}

		}
		catch (IOException ex) {
			System.err.println("Could not read relevant parameters from the config file " + CONFIG_FILE);
			ex.printStackTrace();
			System.exit(1);
		}
		
		finally {
			if (input != null) {
				try {
					input.close();
				}
				catch (IOException e) {
					e.printStackTrace();
					System.exit(1);
				}
			}
		}
		
		if (Settings.PREDICTION_TYPE.equals("aRx")) {
			

			
			int[] values = getMultiProcessing(Settings.PATH_RULES);
			
			PrintWriter log = null;
			
			
			if (values[0] == 0) log = new PrintWriter(Settings.PATH_RULES + "_plog");
			else log = new PrintWriter(Settings.PATH_OUTPUT.replace("|", "") + "_plog");
			
			log.println("Logfile");
			log.println("~~~~~~~\n");
			log.println();
			log.println(IOHelper.getParams());
			log.flush();
			
			for (int value : values) {
				
				long startTime = System.currentTimeMillis();
				
				// long indexStartTime = System.currentTimeMillis();
				String path_output_used = null;
				String path_rules_used = null;
				if (value == 0) {
					path_output_used = Settings.PATH_OUTPUT;
					path_rules_used = Settings.PATH_RULES;
				}
				if (value > 0) {
					path_output_used = Settings.PATH_OUTPUT.replaceFirst("\\|.*\\|", "" + value);
					path_rules_used = Settings.PATH_RULES.replaceFirst("\\|.*\\|", "" + value);
					
				}
				log.println("rules:   " + path_rules_used);
				log.println("output: " + path_output_used);
				log.flush();
			
				

				

				
				PrintWriter pw = new PrintWriter(new File(path_output_used));

				
				if (Settings.PATH_EXPLANATION != null) Settings.EXPLANATION_WRITER = new PrintWriter(new File(Settings.PATH_EXPLANATION));
				System.out.println("* writing prediction to " + path_output_used);
				
				TripleSet trainingSet = new TripleSet(Settings.PATH_TRAINING);
				TripleSet testSet = new TripleSet(Settings.PATH_TEST);
				TripleSet validSet = new TripleSet(Settings.PATH_VALID);
				
				RuleReader rr = new RuleReader();
				LinkedList<Rule> rules = rr.read(path_rules_used);
				/*
				ArrayList<Rule> rules = new ArrayList<Rule>();
				 
				if (FILTER) {
					rules = RuleFilter.filter(rulesBefore, log);
				}
				else {
					rules = rulesBefore;
				}
				*/
				System.out.println("* read rules " + rules.size() + " from file");
				// showRulesStats(rules);
				//System.out.println("* loaded rules = " + rulesBefore.size());
				// showRulesStats(rulesBefore);
				
				
				int rulesSize = rules.size();
				LinkedList<Rule> rulesThresholded = new LinkedList<Rule>();
				if (Settings.THRESHOLD_CONFIDENCE > 0.0) {
					for (Rule r : rules) {
						if (r.getAppliedConfidence() > Settings.THRESHOLD_CONFIDENCE) {
							rulesThresholded.add(r);
						}
					}
					System.out.println("* applied confidence threshold of " + Settings.THRESHOLD_CONFIDENCE + " and reduced from " + rules.size() + " to " + rulesThresholded.size() + " rules");
				}
				rules = rulesThresholded;
				RuleEngine.applyRulesARX(rules, testSet, trainingSet, validSet, Settings.TOP_K_OUTPUT, pw);
			
			
				long endTime = System.currentTimeMillis();
				System.out.println("* evaluated " + rulesSize + " rules to propose candiates for " + testSet.getTriples().size() + "*2 completion tasks");
				System.out.println("* finished in " + (endTime - startTime) + "ms.");
				
				
				System.out.println();
				
				log.println("finished in " + (endTime - startTime) / 1000 + "s.");
				log.println();
				log.flush();
			}
			log.close();
		}
		/*
		if (PREDICTION_TYPE.equals("xRy")) {
		
			
			TripleSet trainingSet = new TripleSet(PATH_TRAINING);
			TripleSet testSet = new TripleSet(PATH_TEST);
			TripleSet validSet = new TripleSet(PATH_VALID);
			
			RuleReader rr = new RuleReader();
			List<Rule> rules = rr.read(PATH_RULES);
			long startTime = System.currentTimeMillis();
			RuleEngine.applyRulesXRY(rules, testSet, trainingSet, validSet, TOP_K_OUTPUT);
		
		
			long endTime = System.currentTimeMillis();
			System.out.println("* evaluated " + rules.size() + " rules to propose candiates for pairwise completion tasks");
			System.out.println("* finished in " + (endTime - startTime) + "ms.");
		}
		*/

		
	
	}

	
	private static void showRulesStats(List<Rule> rules) { 
		int xyCounter = 0;
		int xCounter = 0;
		int yCounter = 0;
		for (Rule rule : rules) {
			if (rule.isXYRule()) xyCounter++;
			if (rule.isXRule()) xCounter++;
			if (rule.isYRule()) yCounter++;
		}
		System.out.println("XY=" + xyCounter + " X="+ xCounter + " Y=" + yCounter);
		
	}

	private static int[] getMultiProcessing(String path1) {
		String token[] = path1.split("\\|");
		if (token.length < 2) {
			return new int[] {0};
		}
		else {
			String numbers[] = token[1].split(",");
			int[] result = new int[numbers.length];
			int i = 0;
			for (String n : numbers) {
				result[i] = Integer.parseInt(n);
				i++;
			}
			return result;
		}
	}
	

		

}
