package de.unima.ki.anyburl;

import java.io.File;
import java.io.FileInputStream;
import java.io.IOException;
import java.io.InputStream;
import java.io.PrintWriter;
import java.text.DecimalFormat;
import java.util.ArrayList;
import java.util.HashSet;
import java.util.LinkedList;
import java.util.Properties;

import java.util.Random;

import de.unima.ki.anyburl.algorithm.PathSampler;
import de.unima.ki.anyburl.data.Triple;
import de.unima.ki.anyburl.data.TripleSet;
import de.unima.ki.anyburl.io.IOHelper;
import de.unima.ki.anyburl.io.RuleReader;
import de.unima.ki.anyburl.structure.Path;
import de.unima.ki.anyburl.structure.RefinedRule;
import de.unima.ki.anyburl.structure.Rule;
import de.unima.ki.anyburl.threads.RScorer;
import de.unima.ki.anyburl.threads.Scorer;

public class Refine {
	
	private static long timeStamp = 0;
	
	private static String CONFIG_FILE = "config-refine.properties";
	
	private static LinkedList<RefinedRule> rulesToBeScored = new LinkedList<RefinedRule>();
	private static LinkedList<RefinedRule> rulesScored = new LinkedList<RefinedRule>();
	private static boolean finished = false;
	



	public static void main(String[] args) throws IOException {
		
		if (args.length == 1) {
			CONFIG_FILE = args[0];
			System.out.println("reading params from file " + CONFIG_FILE);
		}
		
		Properties prop = new Properties();
		InputStream input = null;
		
		try {
			input = new FileInputStream(CONFIG_FILE);
			prop.load(input);
			Settings.PATH_TRAINING = IOHelper.getProperty(prop, "PATH_TRAINING",Settings.PATH_TRAINING);
			Settings.PATH_RULES = IOHelper.getProperty(prop, "PATH_RULES", Settings.PATH_RULES);
			Settings.PATH_OUTPUT = IOHelper.getProperty(prop, "PATH_OUTPUT", Settings.PATH_OUTPUT);	
			Settings.SNAPSHOTS_AT = IOHelper.getProperty(prop, "SNAPSHOTS_AT", Settings.SNAPSHOTS_AT);
			Settings.SAMPLE_SIZE = IOHelper.getProperty(prop, "SAMPLE_SIZE", Settings.SAMPLE_SIZE);
			Settings.TRIAL_SIZE = IOHelper.getProperty(prop, "TRIAL_SIZE", Settings.TRIAL_SIZE);
			Settings.BATCH_TIME = IOHelper.getProperty(prop, "BATCH_TIME", Settings.BATCH_TIME);
			Settings.SATURATION = IOHelper.getProperty(prop, "SATURATION", Settings.SATURATION);
			Settings.WORKER_THREADS = IOHelper.getProperty(prop, "WORKER_THREADS", Settings.WORKER_THREADS);
			Settings.THRESHOLD_CORRECT_PREDICTIONS = IOHelper.getProperty(prop, "THRESHOLD_CORRECT_PREDICTIONS", Settings.THRESHOLD_CORRECT_PREDICTIONS);
			Settings.THRESHOLD_CONFIDENCE = IOHelper.getProperty(prop, "THRESHOLD_CONFIDENCE", Settings.THRESHOLD_CONFIDENCE);
			Settings.UNSEEN_NEGATIVE_EXAMPLES = IOHelper.getProperty(prop, "UNSEEN_NEGATIVE_EXAMPLES", Settings.UNSEEN_NEGATIVE_EXAMPLES);

		}
		catch (IOException ex) {
			System.err.println("Could not read relevant parameters from the config file " + CONFIG_FILE);
			ex.printStackTrace();
			System.exit(1);
		}
		
		finally {
			if (input != null) {
				try {
					input.close();
				}
				catch (IOException e) {
					e.printStackTrace();
					System.exit(1);
				}
			}
		}
		
		PrintWriter log = new PrintWriter(Settings.PATH_OUTPUT + "_log");
		log.println("Logfile");
		log.println("~~~~~~~\n");
		
		long indexStartTime = System.currentTimeMillis();
		
		TripleSet ts = new TripleSet(Settings.PATH_TRAINING);
		
		RuleReader rr = new RuleReader();
		LinkedList<Rule> rules = rr.read(Settings.PATH_RULES, RuleReader.TYPE_CYCLIC);
		
		System.out.println("* loaded " + rules.size() + " cyclic rules to be refined");
		PathSampler ps = new PathSampler(ts);		
		
		
		ArrayList<RefinedRule> refinedRules = new ArrayList<RefinedRule>();
		
		int pathCounter = 0;
		// int batchPathCounter = 0;
		int batchCounter = 0;


		int batchPreviouslyFoundRules = 0;
		int batchNewUsefulRules = 0;
		int batchRules = 0;
		
		Random rand = new Random();
		
		
		int snapshotIndex = 0;
		
		long indexEndTime = System.currentTimeMillis();
		log.println("indexing dataset: " + Settings.PATH_TRAINING);
		log.println("time elapsed: " + (indexEndTime - indexStartTime) + "ms");
		log.println();
		log.println(IOHelper.getParams());
		log.flush();
		
		long startTime = System.currentTimeMillis();
		
		// Thread[] scorer = new Thread[Learn.WORKER_THREADS];
		
		Thread[] rscorer = new Thread[Settings.WORKER_THREADS];
		for (int threadCounter = 0; threadCounter < Settings.WORKER_THREADS; threadCounter++) {
			System.out.println("* creating worker thread #" + threadCounter);
			rscorer[threadCounter] = new RScorer(ts);
			rscorer[threadCounter].start();
		}
		
		
		boolean done = false;
		while (true) {
			long time = System.currentTimeMillis();
			int elapsedSeconds = (int)(time - startTime) / 1000;
			synchronized(Refine.class) {
				if (elapsedSeconds > Settings.SNAPSHOTS_AT[snapshotIndex] || done) {
					storeRules(Settings.PATH_OUTPUT, done ? 0 : Settings.SNAPSHOTS_AT[snapshotIndex], refinedRules, log, elapsedSeconds);
					snapshotIndex++;
					if (!done) System.out.println("CREATED SNAPSHOT " + snapshotIndex + " after " + elapsedSeconds + " seconds");
					else System.out.println("CREATED FINAL SNAPSHOT 0 after " + elapsedSeconds + " seconds");
					if (snapshotIndex == Settings.SNAPSHOTS_AT.length || done) {
						log.flush();
						log.close();
						System.out.println("Bye, bye");
						Refine.finished = true;
						try {
							Thread.sleep(500);
						}
						catch (InterruptedException e) {
							e.printStackTrace();
						}
						System.exit(1);
					}
				}
			}
			long batchStartTime = System.currentTimeMillis();
			long currentTime = 0;
		
			while (true) {
				currentTime = System.currentTimeMillis();
				// System.out.println("currentTime: " + currentTime);
				if (currentTime- batchStartTime > Settings.BATCH_TIME) {
					// System.out.println("batch break");
					break;
				}
				pathCounter++;
				
				Rule motherRule = rules.get(rand.nextInt(rules.size()));
				Triple predictedValidTriple = motherRule.getRandomValidPrediction(ts);
				
				
				Path p = ps.samplePath(2, false, predictedValidTriple);
				
				
				
				if (predictedValidTriple != null && p != null && p.isValid() && p.isNonCyclic(predictedValidTriple)) {

					Rule pr = new Rule(p);
					HashSet<Rule> grules = pr.getGeneralizations(false);					
				
					for (Rule r : grules) {
						// System.out.println("main loop: " + r);
						if (r.isXYRule()) continue;
						if (r.isTrivial()) continue;
						
						RefinedRule refinedRule = new RefinedRule(motherRule, r);
						
						// if (!mineCyclicNotAcyclic && r.isXYRule()) continue;
						
						batchRules++;
						if (!refinedRules.contains(refinedRule)) {
							if (!addRuleToBeScored(refinedRule)) {
								try {
									// System.out.println("before sleep 25");
									Thread.sleep(25);
									
								}
								catch (InterruptedException e) {  e.printStackTrace(); }	
							}
							RefinedRule scoredRule = getScoredRule();
							// System.out.println("retrieving a scored rule: " + scoredRule);
							while (scoredRule != null) {
								if (scoredRule.outperformsMother() && scoredRule.getAppliedConfidence() >= Settings.THRESHOLD_CONFIDENCE &&  scoredRule.getCorrectlyPredicted() >= Settings.THRESHOLD_CORRECT_PREDICTIONS) {
									// System.out.println("  refined rule outperforms");
									if (refinedRules.add(scoredRule)) batchNewUsefulRules++;
								}
								scoredRule = getScoredRule();
							}
						}
						else batchPreviouslyFoundRules++;
					}
				}
			}
			synchronized(Refine.class) {
				batchCounter++;
				double currentCoverage = batchPreviouslyFoundRules / (double)(batchNewUsefulRules + batchPreviouslyFoundRules);
				System.out.println();
				System.out.println(">>> batch #" + batchCounter + " (sampled " + pathCounter + " pathes) *****");
				System.out.println(">>> fraction of previously seen rules within useful rules in this batch: " +  currentCoverage + "");
				System.out.println(">>> new rules above thresholds within this thread: " + batchNewUsefulRules);
				System.out.println(">>> rules found previously or a second time in this batch: " + batchPreviouslyFoundRules);
				System.out.println(">>> all rules created during within this batch: " + batchRules);
				System.out.println(">>> all rules stored so far: " + refinedRules.size());

				if (currentCoverage > Settings.SATURATION || batchNewUsefulRules == 0) {
					done = true;
				}
				
				batchNewUsefulRules = 0;
				batchRules = 0;
				batchPreviouslyFoundRules = 0;
	
				Refine.rulesScored.clear();
				Refine.rulesToBeScored.clear();
			}
		}
		
	}
	
	
	public static synchronized RefinedRule getNextRuleToBeScored()  {
		// System.out.println("poll from todo list");
		return rulesToBeScored.poll();
	}
	
	public static synchronized RefinedRule getScoredRule()  {
		// System.out.println("poll from scored list");
		return rulesScored.poll();
	}
	
	public static synchronized boolean addRuleToBeScored(RefinedRule rule)  {
		if (rulesToBeScored.size() > 10000) {
			// System.out.println("try to add to todo but do not add because list is too large");
			return false;
		}
		else {
			rulesToBeScored.add(rule);
			// System.out.println("add rule to todo list");
			return true;
		}
	}
	
	
	public static synchronized boolean addScoredRule(RefinedRule rule)  {
		// System.out.println("add rule to scored list");
		rulesScored.add(rule);
		if (rulesScored.size() > 10000) {
			// System.out.println("scored list is already very full");
			return false;
		}
		else {
			return true;
		}
	}
	

	


	private static void storeRules(String path, int snapshotCounter, ArrayList<RefinedRule> rules, PrintWriter log, int elapsedSeconds) {
		DecimalFormat df = new DecimalFormat("###.#");
		File ruleFile = new File(path + "-" + snapshotCounter);
		try {
			log.println();
			log.println("refined rule file: " + ruleFile.getPath());
			System.out.println(">>> Storing " + rules.size() + " rules in file " + ruleFile.getPath());
			PrintWriter pw = new PrintWriter(ruleFile);

			for (RefinedRule r : rules) {
				pw.println(r);
			}
			pw.flush();
			pw.close();
			log.println("time planned: " + snapshotCounter +  "s");
			log.println("time elapsed: " + elapsedSeconds + "s");
			log.println("");
			log.flush();
		} catch (IOException e) {
			e.printStackTrace();
		}
		
	}
	
	
	public static boolean finished() {
		return finished;
	}
	

	public static void showElapsedMoreThan(long duration, String message) {
		long now = System.currentTimeMillis();
		long elapsed = now - timeStamp;
		if (elapsed > duration) {
			System.err.println(message + " required " + elapsed + " millis!");
		}
		
	}

	public static void takeTime() {
		timeStamp = System.currentTimeMillis();
		
	}

}
