package de.unima.ki.anyburl;


import java.io.PrintWriter;
import java.text.DecimalFormat;
import java.util.ArrayList;
import java.util.Collections;
import java.util.HashMap;
import java.util.HashSet;
import java.util.LinkedHashMap;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import java.util.Map.Entry;
import java.util.Set;
import java.util.AbstractMap.SimpleImmutableEntry;
import java.util.stream.Collectors;

import de.unima.ki.anyburl.data.*;
import de.unima.ki.anyburl.structure.*;
import de.unima.ki.anyburl.threads.Predictor;

public class RuleEngine {

	private final static double EPSILON = 0.0001;
	
	
	private static LinkedList<Triple> predictionTasks = new LinkedList<Triple>();
	private static int predictionsMade = 0;
	private static PrintWriter predictionsWriter = null;
	
	public static void applyRulesARX(LinkedList<Rule> rules, TripleSet testSet, TripleSet trainingSet, TripleSet validationSet, int k, PrintWriter resultsWriter) {
		System.out.println("* applying rules");
		// HashMap<String, HashSet<Rule>> relation2Rules = createRuleIndex(rules);

		HashMap<String, ArrayList<Rule>> relation2Rules = createOrderedRuleIndex(rules);
		
		System.out.println("* set up index structure covering rules for " + relation2Rules.size() + " different relations");
		TripleSet filterSet = new TripleSet();
		// TODO
		filterSet.addTripleSet(trainingSet);
		filterSet.addTripleSet(validationSet);
		filterSet.addTripleSet(testSet);
		System.out.println("* constructed filter set with " + filterSet.getTriples().size() + " triples");
		if (filterSet.getTriples().size() == 0) {
			System.err.println("WARNING: using empty filter set!");
		}
		// prepare the data structures used a s cache for question that are reoccuring
		// HashMap<SimpleImmutableEntry<String, String>, LinkedHashMap<String, Double>> headCandidateCache = new HashMap<SimpleImmutableEntry<String, String>, LinkedHashMap<String, Double>>();
		// HashMap<SimpleImmutableEntry<String, String>, LinkedHashMap<String, Double>> tailCandidateCache = new HashMap<SimpleImmutableEntry<String, String>, LinkedHashMap<String, Double>>();
		// start iterating over the test cases
		
		// int counter = 0;
		// long startTime = System.currentTimeMillis();
		// long currentTime = 0;
		
		
		ScoreTree.LOWER_BOUND = k;
		ScoreTree.UPPER_BOUND = ScoreTree.LOWER_BOUND;
		ScoreTree.EPSILON = EPSILON;

		
		predictionTasks.addAll(testSet.getTriples());
		predictionsWriter = resultsWriter;
		
		Thread[] predictors = new Thread[Settings.WORKER_THREADS];
		for (int threadCounter = 0; threadCounter < Settings.WORKER_THREADS; threadCounter++) {
			System.out.println("* creating worker thread #" + threadCounter);
			predictors[threadCounter] = new Predictor(testSet, trainingSet, filterSet, k, relation2Rules);
			predictors[threadCounter].start();
		}
		
		
		
		while (alive(predictors)) {
			try {
				Thread.sleep(500);
			}
			catch (InterruptedException e) {
				e.printStackTrace();
			}
		}
		predictionsWriter.flush();
		predictionsWriter.close();
		
		System.out.println("* done with rule application");
		RuleEngine.predictionsMade = 0;

	}
	
	private static boolean alive(Thread[] threads) {
		for (Thread t : threads) {
			if (t.isAlive()) return true;
		}
		return false;
	}
	
	
	
	public static synchronized Triple getNextPredictionTask()  {
		predictionsMade++;
		Triple triple = predictionTasks.poll();
		if (predictionsMade % 500 == 0) {
			System.out.println("* (#" + predictionsMade + ") trying to guess the tail and head of " + triple.toString());
			predictionsWriter.flush();
		}
		return triple;
	}

	

	public static void predictMax(TripleSet testSet, TripleSet trainingSet, TripleSet filterSet, int k, HashMap<String, ArrayList<Rule>> relation2Rules, Triple triple) {

		String relation = triple.getRelation();
		String head = triple.getHead();
		String tail = triple.getTail();

		ScoreTree kTailTree = new ScoreTree();
		ScoreTree kHeadTree = new ScoreTree();

		if (relation2Rules.containsKey(relation)) {
			ArrayList<Rule> relevantRules = relation2Rules.get(relation);
			for (Rule rule : relevantRules) {
				if (!kTailTree.fine()) {
					HashSet<String> tailCandidates = rule.computeTailResults(head, trainingSet);
					HashSet<String> fTailCandidates = getFilteredEntities(filterSet, testSet, triple, tailCandidates, true);
					if (Apply.FB15_237_TRICK) fTailCandidates = applyFB15237TrickTail(triple,fTailCandidates, trainingSet);
					kTailTree.addValues(rule.getAppliedConfidence(), fTailCandidates);
				}
				else {
					break;
				}
			}
			for (Rule rule : relevantRules) {
				
				if (!kHeadTree.fine()) {
					HashSet<String> headCandidates = rule.computeHeadResults(tail, trainingSet);
					HashSet<String> fHeadCandidates = getFilteredEntities(filterSet, testSet, triple, headCandidates, false);
					if (Apply.FB15_237_TRICK) fHeadCandidates = applyFB15237TrickHead(triple,fHeadCandidates, trainingSet);
					kHeadTree.addValues(rule.getAppliedConfidence(), fHeadCandidates);
				}
				else {
					break;
				}				
			}
		}
		
		LinkedHashMap<String, Double> kTailCandidates = new LinkedHashMap<String, Double>();
		LinkedHashMap<String, Double> kHeadCandidates = new LinkedHashMap<String, Double>();
		
		kTailTree.getAsLinkedList(kTailCandidates);
		kHeadTree.getAsLinkedList(kHeadCandidates);

		// final sorting
		kTailCandidates = sortByValue(kTailCandidates);
		kHeadCandidates = sortByValue(kHeadCandidates);

		writeTopKCandidates(triple, testSet, kHeadCandidates, kTailCandidates, predictionsWriter, k);
	}
	
	
	public static void predictNoisyOr(TripleSet testSet, TripleSet trainingSet, TripleSet filterSet, int k, HashMap<String, ArrayList<Rule>> relation2Rules, Triple triple) {

		String relation = triple.getRelation();
		String head = triple.getHead();
		String tail = triple.getTail();
		
		HashMap<String, HashSet<Rule>> allTailCandidates = new HashMap<String, HashSet<Rule>>();
		HashMap<String, HashSet<Rule>> allHeadCandidates = new HashMap<String, HashSet<Rule>>();
		
		if (relation2Rules.containsKey(relation)) {
			ArrayList<Rule> relevantRules = relation2Rules.get(relation);
			for (Rule rule : relevantRules) {
				
				HashSet<String> tailCandidates = rule.computeTailResults(head, trainingSet);
				HashSet<String> fTailCandidates = getFilteredEntities(filterSet, testSet, triple, tailCandidates, true);
				for (String fTailCandidate : fTailCandidates) {
					if (!allTailCandidates.containsKey(fTailCandidate)) allTailCandidates.put(fTailCandidate, new HashSet<Rule>());
					allTailCandidates.get(fTailCandidate).add(rule);
				}
				
				HashSet<String> headCandidates = rule.computeHeadResults(tail, trainingSet);
				HashSet<String> fHeadCandidates = getFilteredEntities(filterSet, testSet, triple, headCandidates, false);
				for (String fHeadCandidate : fHeadCandidates) {
					if (!allHeadCandidates.containsKey(fHeadCandidate)) allHeadCandidates.put(fHeadCandidate, new HashSet<Rule>());
					allHeadCandidates.get(fHeadCandidate).add(rule);
				}
				
				
			}

		}
		
		LinkedHashMap<String, Double> kTailCandidates = new LinkedHashMap<String, Double>();
		LinkedHashMap<String, Double> kHeadCandidates = new LinkedHashMap<String, Double>();
	
		computeNoisyOr(allTailCandidates, kTailCandidates);
		computeNoisyOr(allHeadCandidates, kHeadCandidates);
		// final sorting
		kTailCandidates = sortByValue(kTailCandidates);
		kHeadCandidates = sortByValue(kHeadCandidates);
	
		
		if (Settings.PATH_EXPLANATION != null) writeTopKCandidatesPlusExplanation(triple, testSet, kHeadCandidates, allHeadCandidates, kTailCandidates, allTailCandidates, predictionsWriter, k);
		else writeTopKCandidates(triple, testSet, kHeadCandidates, kTailCandidates, predictionsWriter, k);

	}
	
	private static void computeNoisyOr(HashMap<String, HashSet<Rule>> allCandidates, LinkedHashMap<String, Double> kCandidates) { 
		for (String cand : allCandidates.keySet()) {
			double cp = 1.0;
			for (Rule r : allCandidates.get(cand)) {
				cp = cp * (1.0 - r.getAppliedConfidence());
			}
			kCandidates.put(cand, 1.0 - cp);
		}
		
	}
	
	private static HashSet<String> applyFB15237TrickTail(Triple triple, HashSet<String> tailCandidates, TripleSet trainingSet) {
		if (tailCandidates == null) return null; 
		HashSet<String> filtered = new HashSet<String>();
		for (String cand : tailCandidates) {
			if (trainingSet.getRelations(triple.getHead(), cand).size() == 0 && trainingSet.getRelations(cand, triple.getHead()).size() == 0) {
				filtered.add(cand);
			}
		}
		return filtered;
	}
	
	private static HashSet<String> applyFB15237TrickHead(Triple triple, HashSet<String> headCandidates, TripleSet trainingSet) {
		if (headCandidates == null) return null; 
		HashSet<String> filtered = new HashSet<String>();
		for (String cand : headCandidates) {
			if (trainingSet.getRelations(cand, triple.getTail()).size() == 0 && trainingSet.getRelations(triple.getTail(), cand).size() == 0) {
				filtered.add(cand);
			}
		}
		return filtered;
	}


	private static void show(LinkedHashMap<String, Double> kCandidates, String headline) {
		System.out.println("*** " + headline + " ***");
		for (String candidate : kCandidates.keySet()) {
			double conf = kCandidates.get(candidate);
			System.out.println(conf + " = " + candidate);
		}
	}

	private static HashMap<String, HashSet<Rule>> createRuleIndex(List<Rule> rules) {
		
		int counterL1C = 0;
		int counterL2C = 0;
		int counterL1AC = 0;
		int counterL1AN = 0;
		int counterOther = 0;
		
		HashMap<String, HashSet<Rule>> relation2Rules = new HashMap<String, HashSet<Rule>>();
		for (Rule rule : rules) {
			
			/*
			
			if (rule.isXYRule()) {
				if (rule.bodysize() == 1)  counterL1C++;
				if (rule.bodysize() == 2)  counterL2C++;
			}
			else {
				
				if (rule.bodysize() == 1)  {
					if (rule.hasConstantInBody()) counterL1AC++;
					else counterL1AN++;
				}
				else {
					if (rule.hasConstantInBody()) continue;
				}	
			}
			*/
			
			String relation = rule.getTargetRelation();
			if (!relation2Rules.containsKey(relation)) relation2Rules.put(relation, new HashSet<Rule>());
			relation2Rules.get(relation).add(rule);
			
			
		}
		// System.out.println("L1C=" + counterL1C + " L2C=" + counterL2C + " L1AC=" + counterL1AC + " L1AN=" + counterL1AN + " OTHER=" + counterOther);
		return relation2Rules;
	}
	
	
	private static HashMap<String, ArrayList<Rule>> createOrderedRuleIndex(LinkedList<Rule> rules) {
		HashMap<String, ArrayList<Rule>> relation2Rules = new HashMap<String, ArrayList<Rule>>();
		long l = 0;
		while (rules.size() > 0) {
			Rule rule = rules.poll();
			// Rule rule = rules.remove(rules.size()-1);
			String relation = rule.getTargetRelation();
			
			if (!relation2Rules.containsKey(relation)) {
				relation2Rules.put(relation, new ArrayList<Rule>());
			}
			relation2Rules.get(relation).add(rule);	
			if (l % 100000 == 0) {
				// rules.trimToSize();
				System.out.println("* indexed " + l + " rules");
			}
			l++;
		}
		
		for (String relation : relation2Rules.keySet()) {
			// System.out.println("* sort " + relation2Rules.get(relation).size() + "  rules for relation " + relation );
			relation2Rules.get(relation).trimToSize();
			relation2Rules.get(relation).sort(new RuleConfidenceComparator());
		}
		return relation2Rules;
	}


	/*
	private static void updateCandidateProbabailities(Rule rule, boolean tailNotHead, String candidate, HashMap<String, Double> candidates2Probabilities) {
		double prob = rule.getAppliedConfidence();
		if (!candidates2Probabilities.containsKey(candidate)) candidates2Probabilities.put(candidate, prob);
		else {
			double previousProb = candidates2Probabilities.get(candidate);
			double newProb = combineProbability(prob, previousProb);
			candidates2Probabilities.put(candidate, newProb);
		}
	}
	*/
	
	
	private static LinkedHashMapK getFilteredCandidates(TripleSet filterSet, TripleSet testSet, Triple t, HashMap<String, Double> candidates, boolean tailNotHead) {
		// LinkedHashMap<String, Double> candidatesSorted = sortByValue(candidates);
		LinkedHashMap<String, Double> kCandidates = new LinkedHashMap<String, Double>();
		int i = 0;
		for (Entry<String, Double> entry : candidates.entrySet()) {
			if (!tailNotHead) {
				if (!filterSet.isTrue(entry.getKey(), t.getRelation(), t.getTail())) {
					kCandidates.put(entry.getKey(), entry.getValue());
					i++;
				}
				if (testSet.isTrue(entry.getKey(), t.getRelation(), t.getTail())) {
					kCandidates.put(entry.getKey(), entry.getValue());
				}
			}
			if (tailNotHead) {
				if (!filterSet.isTrue(t.getHead(), t.getRelation(), entry.getKey())) {
					kCandidates.put(entry.getKey(), entry.getValue());
					i++;
				}
				if (testSet.isTrue(t.getHead(), t.getRelation(), entry.getKey())) {
					kCandidates.put(entry.getKey(), entry.getValue());
				}
			}
		}
		return (new LinkedHashMapK(kCandidates, i));
	}
	
	
	
	private static HashSet<String> getFilteredEntities(TripleSet filterSet, TripleSet testSet, Triple t, HashSet<String> candidateEntities, boolean tailNotHead) {
		// LinkedHashMap<String, Double> candidatesSorted = sortByValue(candidates);
		HashSet<String> filteredEntities = new HashSet<String>();
		for (String entity : candidateEntities) {
			if (!tailNotHead) {
				if (!filterSet.isTrue(entity, t.getRelation(), t.getTail())) {
					filteredEntities.add(entity);
				}
				
				if (testSet.isTrue(entity, t.getRelation(), t.getTail())) {
					// TAKE CARE, remove to reactivate the possibility of storing previous results
					if (entity.equals(t.getHead())) filteredEntities.add(entity);
				}
				
			}
			if (tailNotHead) {
				if (!filterSet.isTrue(t.getHead(), t.getRelation(), entity)) {
					filteredEntities.add(entity);
				}
				if (testSet.isTrue(t.getHead(), t.getRelation(), entity)) {
					// TAKE CARE, remove to reactivate the possibility of storing previous results
					if (entity.equals(t.getTail())) filteredEntities.add(entity);
				}
			}
		}
		return filteredEntities;
	}
	
	private static synchronized void writeTopKCandidates(Triple t, TripleSet testSet, LinkedHashMap<String, Double> kHeadCandidates, LinkedHashMap<String, Double> kTailCandidates, PrintWriter writer, int k) {
		writer.println(t);
		int i = 0;
		writer.print("Heads: ");
		for (Entry<String, Double> entry : kHeadCandidates.entrySet()) {
			if (t.getHead().equals(entry.getKey()) || !testSet.isTrue(entry.getKey(), t.getRelation(), t.getTail())) {
				writer.print(entry.getKey() + "\t" + entry.getValue() + "\t");
				i++;
			}
			if (i == k) break;
		}
		writer.println();
		i = 0;
		writer.print("Tails: ");
		for (Entry<String, Double> entry : kTailCandidates.entrySet()) {
			if (t.getTail().equals(entry.getKey()) || !testSet.isTrue(t.getHead(), t.getRelation(), entry.getKey())) {
				writer.print(entry.getKey() + "\t" + entry.getValue() + "\t");
				i++;
			}
			if (i == k) break;
		}
		writer.println();
		writer.flush();

	}
	
	private static synchronized void writeTopKCandidatesPlusExplanation(Triple t, TripleSet testSet, LinkedHashMap<String, Double> kHeadCandidates, HashMap<String, HashSet<Rule>> allHeadCandidates, LinkedHashMap<String, Double> kTailCandidates, HashMap<String, HashSet<Rule>> allTailCandidates, PrintWriter writer, int k) {
		writer.println(t);
		Settings.EXPLANATION_WRITER.println(t);
		int i = 0;
		writer.print("Heads: ");
		Settings.EXPLANATION_WRITER.println("Heads:");
		for (Entry<String, Double> entry : kHeadCandidates.entrySet()) {
			if (t.getHead().equals(entry.getKey()) || !testSet.isTrue(entry.getKey(), t.getRelation(), t.getTail())) {
				writer.print(entry.getKey() + "\t" + entry.getValue() + "\t");
				Settings.EXPLANATION_WRITER.println("\t" + entry.getKey() + "\t" + entry.getValue() + "\t");
				
				
				ArrayList<Rule> sortedRules = new ArrayList<Rule>();
				sortedRules.addAll(allHeadCandidates.get(entry.getKey()));
				sortedRules.sort((r1, r2) -> Double.compare(r1.getAppliedConfidence(), r2.getAppliedConfidence()));
				
				for (Rule r : sortedRules) {
					Settings.EXPLANATION_WRITER.println("\t\t" + r);
				}
				i++;
			}
			if (i == k) break;
		}
		writer.println();
		i = 0;
		writer.print("Tails: ");
		Settings.EXPLANATION_WRITER.println("Tails:");
		for (Entry<String, Double> entry : kTailCandidates.entrySet()) {
			if (t.getTail().equals(entry.getKey()) || !testSet.isTrue(t.getHead(), t.getRelation(), entry.getKey())) {
				writer.print(entry.getKey() + "\t" + entry.getValue() + "\t");
				Settings.EXPLANATION_WRITER.println("\t" + entry.getKey() + "\t" + entry.getValue() + "\t");
				for (Rule r : allTailCandidates.get(entry.getKey())) {
					Settings.EXPLANATION_WRITER.println("\t\t" + r);
				}
				i++;
			}
			if (i == k) break;
		}
		Settings.EXPLANATION_WRITER.flush();
		writer.println();
		writer.flush();
	}
	
	/*
	private static void processTopKCandidates(TripleSet testSet, Triple t, HashMap<String, Double> tailCandidates, HashMap<String, Double> headCandidates, TripleSet filterSet, int k, PrintWriter writer, HashMap<String, Double> kTailCandidates, HashMap<String, Double> kHeadCandidates) {
		LinkedHashMap<String, Double> tailCandidatesSorted = sortByValue(tailCandidates);
		LinkedHashMap<String, Double> headCandidatesSorted = sortByValue(headCandidates);
		writer.println(t);
		writer.print("Heads: ");
		int i = 0;
		for (Entry<String, Double> entry : headCandidatesSorted.entrySet()) {
			if (i < k) {
				if (!filterSet.isTrue(entry.getKey(), t.getRelation(), t.getTail()) || t.getHead().equals(entry.getKey())) {
					writer.print(entry.getKey() + "\t" + entry.getValue() + "\t");
					kHeadCandidates.put(entry.getKey(), entry.getValue());
					i++;
				}
				if (testSet.isTrue(entry.getKey(), t.getRelation(), t.getTail())) {
					kHeadCandidates.put(entry.getKey(), entry.getValue());
				}
			}
		}
		writer.println();
		writer.print("Tails: ");
		int j = 0;
		for (Entry<String, Double> entry : tailCandidatesSorted.entrySet()) {
			if (j < k) {
				if (!filterSet.isTrue(t.getHead(), t.getRelation(), entry.getKey())
						|| t.getTail().equals(entry.getKey())) {
					writer.print(entry.getKey() + "\t" + entry.getValue() + "\t");
					kTailCandidates.put(entry.getKey(), entry.getValue());
					j++;
				}
				if (testSet.isTrue(t.getHead(), t.getRelation(), entry.getKey())) {
					kTailCandidates.put(entry.getKey(), entry.getValue());
				}
			}
		}
		writer.println();
		writer.flush();
	}
	*/

	/*
	private static double combineProbability(double prob, double previousProb) {
		double newProb;
		switch (COMBINATION_RULE_ID) {
		case 1: // multiplication
			newProb = 1.0 - ((1.0 - previousProb) * (1.0 - prob));
			break;
		case 2: // maxplus
			newProb = Math.max(previousProb, prob) + EPSILON;
			break;
		case 3: // max
		default:
			newProb = Math.max(previousProb, prob);
			break;
		}
		return newProb;
	}
	*/
	

	
	public static <K, V extends Comparable<? super V>> LinkedHashMap<K, V> sortByValue(Map<K, V> map) {
		return map.entrySet().stream().sorted(Map.Entry.comparingByValue(Collections.reverseOrder()))
				.collect(Collectors.toMap(Map.Entry::getKey, Map.Entry::getValue, (e1, e2) -> e1, LinkedHashMap::new));
	}
	

	private static String f(double d) {
		DecimalFormat df = new DecimalFormat("0.000");
		return df.format(d);
	}

	
}
