package de.unima.ki.anyburl;

import java.io.PrintWriter;

public class Settings {
	
	
	/**
	 * Defines the prediction type which also influences the usage of the other parameters. 
	 * Possible values are currently aRx and xRy.
	 */
	public static String PREDICTION_TYPE = "aRx";
	
	
	/**
	 * Path to the file that contains the triple set used for learning the rules.
	 */
	public static String PATH_TRAINING = "";
	
	/**
	 * Path to the file that contains the triple set used for to test the rules.
	 */
	public static String PATH_TEST = "";
	
	/**
	 * Path to the file that contains the triple set used validation purpose (e.g. learning hyper parameter).
	 */
	public static String PATH_VALID = "";
	
	
	/**
	 * Path to the file that contains the rules that will be refined.
	 */
	public static String PATH_RULES = "";
	
	/**
	 * Path to the output file where the refined rules will be stored.
	 */
	public static String PATH_OUTPUT = "";
	
	/**
	 * Path to the output file where the explanations are stored. If not set no explanations are stored.
	 */
	public static String PATH_EXPLANATION = null;
	
	public static PrintWriter EXPLANATION_WRITER = null;
	
	
	
	/**
	 * Takes a snapshot of the rules refined after each time interval specified in seconds.
	 */
	public static int[] SNAPSHOTS_AT = new int[] {10,100};
	
	
	/**
	* Used for restricting the number of samples drawn for computing scores as confidence.
	*/
	public static int SAMPLE_SIZE = 1000;
	
	
	/**
	* Number of maximal attempts to create body grounding. Every partial body grounding is counted.
	*/
	public static int TRIAL_SIZE = 1000000;
	
	
	/**
	 * Returns only results for head or tail computation if the results set has less elements than this bound.
	 * The idea is that any results set which has more elements is anyhow not useful for a top-k ranking. 
	 * Should be set to a value thats higher than the k of the requested top-k (however, the higher the value,
	 * the more runtime is required)
	 */
	public static int DISCRIMINATION_BOUND = 1000;	

		
	/**
	 * The time that is reserved for one batch in milliseconds. 
	 */
	public static int BATCH_TIME = 1000;
	
	
	/**
	 * The maximal number of body atoms in cyclic rules (inclusive this number). If this number is exceeded all computation time
	 * is used for acyclic rules only from that time on.
	 * 
	 */
	public static int MAX_LENGTH_CYCLIC = 3;
	
	
	/**
	 * The maximal number of body atoms in acyclic rules (inclusive this number). If this number is exceeded all computation time
	 * is used for cyclic rules only from that time on.
	 * 
	 */
	public static int MAX_LENGTH_ACYCLIC = 2;
	
	
	/**
	 * The saturation defined when to stop the refinement process. Once the saturation is reached, no further refinements are searched for.
	 */
	public static double SATURATION = 0.99;
	
	/**
	 * The threshold for the number of correct prediction created with the refined rule.
	 */
	public static int THRESHOLD_CORRECT_PREDICTIONS = 2;
	
	/**
	 * The number of negative examples for which we assume that they exist, however, we have not seen them. Rules with high coverage are favored the higher the chosen number. 
	 */
	public static int UNSEEN_NEGATIVE_EXAMPLES = 5;	

	/**
	 * The threshold for the confidence of the refined rule
	 */
	public static double THRESHOLD_CONFIDENCE = 0.01;
	
	/** 
	 * The number of worker threads which compute the scores of the constructed rules, should be one less then the number of available cores.
	 */
	public static int WORKER_THREADS = 3;
	
	
	/**
	* Defines how to combine probabilities that come from different rules
	* Possible values are: maxplus, noisyor
	*/
	public static String AGGREGATION_TYPE = "maxplus";
	
	public static int AGGREGATION_ID = 1;
	

	/**
	 * The top-k results that are after filtering kept in the results. 
	 */
	public static int TOP_K_OUTPUT = 10;	
	

}
