package de.unima.ki.anyburl.structure;

import java.io.IOException;
import java.text.DecimalFormat;
import java.text.NumberFormat;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Locale;
import java.util.Set;

import de.unima.ki.anyburl.Apply;
import de.unima.ki.anyburl.Learn;
import de.unima.ki.anyburl.Settings;
import de.unima.ki.anyburl.data.SampledPairedResultSet;
import de.unima.ki.anyburl.data.Triple;
import de.unima.ki.anyburl.data.TripleSet;
import de.unima.ki.anyburl.io.RuleReader;

public class RefinedRule {
	
	private Rule motherRule = null;
	private Rule refinedRule = null;
	
	private int hashcode = 0;
	private boolean hashcodeInitialized = false;
	
	// scores
	
	// counted in terms of head groundings
	private int predicted = 0;
	private int correctlyPredicted = 0;
	private double confidence = 0.0; 
	
	// private static DecimalFormat DF = new DecimalFormat("#.####");
	
	private static DecimalFormat DF = (DecimalFormat)NumberFormat.getNumberInstance(new Locale("en", "UK"));
	
	static {
		DF.applyPattern("#.####");	
	}

	
	
	private static final String[] variables = new String[] {"A", "B", "C", "D", "E", "F", "G", "H", "I", "J", "K", "L", "M", "N", "O", "P"};
	
	private int nextFreeVariable = 0;
	
	public RefinedRule(Rule mother, Rule refinement) {
		this.motherRule = mother;
		this.refinedRule = refinement;
	}

	
	public int getPredicted() {
		return this.predicted;
	}
	
	public int getCorrectlyPredicted() {
		return this.correctlyPredicted;
	}


	public double getConfidence() {
		return confidence;
	}
	
	// TODO only quickly fixed, rewrite to make it nicer
	public String toString() {
		StringBuffer sb = new StringBuffer();
		sb.append(this.predicted + "\t");
		sb.append(this.correctlyPredicted + "\t");
		sb.append(this.motherRule.predicted + "\t");
		sb.append(this.motherRule.correctlyPredicted + "\t");
		sb.append(DF.format(this.getAppliedConfidence()) + "<-" + DF.format(this.motherRule.getAppliedConfidence()) + " [" + DF.format(this.confidence) + "<-" + DF.format(this.motherRule.getConfidence()) + "]\t");
		sb.append(this.motherRule.head);
		sb.append(" <= ");
		sb.append(this.motherRule.getBodyString());
		sb.append("; ");
		sb.append(this.refinedRule.getBodyString());
		return sb.toString();
	}
	
	

	public boolean equals(Object thatObject) {
		if (thatObject instanceof RefinedRule) {
			RefinedRule that = (RefinedRule)thatObject;
			if (this.motherRule.equals(that.motherRule) && that.refinedRule.equals(that.refinedRule)) {
				return true;
			}
			if (this.motherRule.equals(that.refinedRule) && that.refinedRule.equals(that.motherRule)) {
				return true;
			}
			return false;
		}
		return false;
	}
	
	public int hashCode() {
		if (!this.hashcodeInitialized) {
			this.hashcode = this.motherRule.hashCode() + this.refinedRule.hashCode();
			this.hashcodeInitialized = true;
		}
		return this.hashcode;
	}

	public void computeScores(TripleSet triples) {
		ArrayList<Triple> motherPredictions = this.motherRule.getRandomPredictions(triples, false);
		this.correctlyPredicted = 0;
		this.predicted = 0;
		for (Triple t : motherPredictions) {
			if (this.refinedRule.isPredictedX(t.getHead(), t.getTail(), triples)) {
				this.predicted++;
				if (triples.isTrue(t)) {
					this.correctlyPredicted++;
				}
			}
		}
		this.confidence = (double)correctlyPredicted / (double)predicted;
	}
	

	public String getTargetRelation() {
		return this.motherRule.head.getRelation();
	}


	public double getAppliedConfidence() {
		double prob = (double)this.getCorrectlyPredicted() / ((double)this.getPredicted() + Settings.UNSEEN_NEGATIVE_EXAMPLES);
		return prob;
	}

	public boolean outperformsMother() {
		if (this.getAppliedConfidence() > this.motherRule.getAppliedConfidence()) return true;
		return false;
	}

}
